@@ -1,5 +1,5 @@
 """
- * $Id: 1.0+8 utils.py a92a642340f2 2021-02-12 francesco.serafin.3@gmail.com $
+ * $Id: 1.0+16 utils.py 52dfe2315c83 2021-05-18 od $
  *
  * This file is part of the Cloud Services Integration Platform (CSIP),
  * a Model-as-a-Service framework, API and application suite.
@@ -15,6 +15,7 @@ from pathlib import Path
 from shutil import make_archive
 from time import sleep
 from typing import List, Dict, Tuple
+from requests.exceptions import Timeout
 
 import json
 import os
@@ -64,12 +65,16 @@ class Client(object):
     STATUS_FAILED = "Failed"
     STATUS_RUNNING = "Running"
     STATUS_QUEUED = "Queued"
+    STATUS_UNKNOWN = "Unknown"
 
     HDR_CSIP_WEBHOOK = "X-CSIP-Webhook"
     HDR_CSIP_REQUESTFILE = "X-CSIP-Request-File"
 
     def __init__(self, data: Dict = None, metainfo: Dict = None,
-                 parent: 'Client' = None, url: str = None, name: str = '', descr: str = ''):
+                 parent: 'Client' = None, url: str = None, name: str = '',
+                 descr: str = '', http_status: int = 0, conf: Dict = None):
+        self.http_status = http_status
+        self.conf = conf or {}
         # parameter data or result data
         self.data = Client.__create_dict(data or [])
         # metainfo 
@@ -99,12 +104,6 @@ class Client(object):
             data = json.load(f)
         return data
 
-    @staticmethod
-    def save_json(file: str, data: Dict) -> None:
-        """Save the file"""
-        with open(file, 'w') as outfile:
-            json.dump(data, outfile)
-
     @staticmethod
     def __create_dict(json: List[Dict]) -> Dict:
         """Create dict key:name, value: whole json"""
@@ -122,21 +121,98 @@ class Client(object):
                 del d[i]
 
     @staticmethod
-    def __get(url: str, section: str = PARAMETER) -> "Client":
+    def __get(url: str, section: str = PARAMETER, conf: Dict = None) -> "Client":
         """HTTP GET to generate a Client"""
-        response = requests.get(url, allow_redirects=True)
-        if response.status_code != requests.codes.ok:
-            response.raise_for_status()
-        response_json = response.json()
+        response, http_status = Client.__get0(url, conf)
+        if http_status == 200:
+            response_json = response.json()
+            data = response_json.get(section, None)
+            mi = response_json.get(Client.METAINFO, {})
+        else:
+            mi = {}
+            data = None
 
-        # print(json_response)
-        # check if suid error happens
-        # This handles  /q/<suid>
-        data = {} if not section in response_json else response_json[section]
-        return Client(data, metainfo=response_json[Client.METAINFO])
+        # if not section in response_json:
+        #     data = {}
+        # else:
+        #     data = response_json[section]
+
+        # data = {} if not section in response_json else response_json[section]
+        # mi = {} if http_status != 200 else response_json[Client.METAINFO]
+        return Client(data, metainfo=mi, http_status=http_status, conf=conf)
 
     @staticmethod
-    def from_file(file) -> "Client":
+    def __get0(url: str, conf: Dict):
+        """HTTP GET to generate a Client"""
+        conf = conf or {}
+        to = conf.get('get_timeout', 4.0)
+        retry = conf.get('retry', 3)
+        allow_redirects = conf.get('allow_redirects', True)
+        http_status = 200
+        response = None
+        while retry > 0:
+            try:
+                response = requests.get(url, allow_redirects=allow_redirects, timeout=to)
+                response.raise_for_status()
+                # response_json = response.json()
+                break
+            except requests.exceptions.HTTPError as errh:
+                http_status = response.status_code
+                # print("Http Error:", errh)
+                # print(response.status_code)
+                break
+            except requests.exceptions.Timeout as errt:
+                ## we will try again
+                retry += -1
+                http_status = -1
+                # print("Timeout Error:", errt)
+            except requests.exceptions.ConnectionError as errc:
+                retry += -1
+                http_status = -2
+                # print("Error Connecting:", errc)
+            except requests.exceptions.RequestException as err:
+                http_status = -3
+                break
+        return response, http_status
+
+    @staticmethod
+    def __post0(url: str, files: Dict, headers: Dict, conf: Dict):
+        """HTTP POST to generate a Client"""
+        conf = conf or {}
+        to = conf.get('post_timeout', 10.0)
+        retry = conf.get('retry', 3)
+        allow_redirects = conf.get('allow_redirects', True)
+        http_status = 200
+        response_json = None
+        while retry > 0:
+            try:
+                response = requests.post(url, files=files, allow_redirects=allow_redirects, headers=headers,
+                                         timeout=to)
+                response.raise_for_status()
+                response_json = response.json()
+                break
+            except requests.exceptions.HTTPError as errh:
+                http_status = response.status_code
+                # print("Http Error:", errh)
+                # print(response.status_code)
+                break
+            except requests.exceptions.Timeout as errt:
+                ## we will try again
+                retry += -1
+                http_status = -1
+                # print("Timeout Error:", errt)
+            except requests.exceptions.ConnectionError as errc:
+                print("Http Connection Error:", errc)
+                retry += -1
+                http_status = -2
+                # print("Error Connecting:", errc)
+            except requests.exceptions.RequestException as err:
+                http_status = -3
+                break
+        return response_json, http_status
+
+    @staticmethod
+    def from_file(file: str) -> "Client":
         """Create a client from a file"""
         d = Client.load_json(file)
         data = d[Client.PARAMETER]
@@ -155,33 +231,47 @@ class Client(object):
         # print(json_response)
         # check if suid error happens
         data = {} if not section in response_json else response_json[section]
-        return Client(data, metainfo=response_json[Client.METAINFO])
+        return Client(data, metainfo=response_json.get(Client.METAINFO, {}))
 
     @staticmethod
-    def get_catalog(url: str) -> List['Client']:
+    def get_catalog(url: str, conf: Dict = None) -> List['Client']:
         """ Get the list of services fron an endpoint"""
 
-        response = requests.get(url, allow_redirects=True)
-        if response.status_code != requests.codes.ok:
-            response.raise_for_status()
+        response, http_status = Client.__get0(url, conf)
+        if http_status != 200:
+            raise Exception("HTTP Error " + http_status)
+
         response_json = response.json()
-        return [Client(metainfo=service) for service in response_json]
+        return [Client(metainfo=service, conf=conf) for service in response_json]
 
     @staticmethod
-    def get_capabilities(c) -> "Client":
+    def get_capabilities(c, conf: Dict = None) -> "Client":
         """Get the parameter of a service. Creates a request Client"""
         if isinstance(c, str):
-            return Client.__get(c)
+            return Client.__get(c, conf=conf)
         if isinstance(c, Client):
             url = c.get_url()
             if url is None:
                 raise Exception("No url info in client argument")
-            cl = Client.__get(url)
+            cl = Client.__get(url, conf=conf)
             c.add_all_metainfo(cl)
             c.add_data(cl)
             return c
         raise Exception("Illegal argument type: " + type(c))
 
+    def save_to(self, file: str) -> None:
+        """Publish/subscribe execution """
+        d = { Client.METAINFO: self.metainfo }
+        if self.parent:
+            d[Client.PARAMETER] = self.parent.data
+            d[Client.RESULT] = self.data
+        else:
+            d[Client.PARAMETER] = self.data
+
+        """Save the file"""
+        with open(file, 'w') as outfile:
+            json.dump(d, outfile)
+
     def pubsub(self, url: str, callback=None, extra_metainfo: Dict = None) -> None:
         """Publish/subscribe execution """
 
@@ -242,22 +332,25 @@ class Client(object):
             if url is None:
                 raise Exception('No url provided.')
         # print(multipart)
+        # response = requests.post(url, files=multipart, allow_redirects=True, headers=headers)
 
-        response = requests.post(url, files=multipart, allow_redirects=True, headers=headers)
+        response_json, http_status = Client.__post0(url, multipart, headers, self.conf)
 
-        if response.status_code != requests.codes.ok:
-            response.raise_for_status()
-
-        response_json = response.json()
+        if http_status == 200:
+            data = response_json.get(Client.RESULT, {})
+            mi = response_json.get(Client.METAINFO, {})
+        else:
+            mi = {}
+            data = {}
 
         # print(response_json)
-        # async handling
-        data = {} if not self.RESULT in response_json else response_json[self.RESULT]
-        return Client(data, parent=self, metainfo=response_json[self.METAINFO])
+        # print(http_status)
+
+        return Client(data, parent=self, metainfo=mi, http_status=http_status, conf=self.conf)
 
     def execute_async(self, url: str = None, files: List[str] = None,
                       first_poll: int = None, next_poll: int = None,
-                      callback=None) -> 'Client':
+                      callback=None, headers: Dict = None) -> 'Client':
         """Executes a service as HTTP/POST. This method runs the service asynchronously and
            returns when the service is done. It 'pings' the service run for completion in the
            frequency provided by 'fist_poll' and 'next_poll'. Provide a callback method to capture
@@ -280,7 +373,7 @@ class Client(object):
         callback = callback or cb
 
         # -> Submitted
-        resp = self.execute(url, files=files, sync=False)
+        resp = self.execute(url, files=files, sync=False, headers=headers)
         callback(resp, "")
 
         fp = (first_poll or resp.get_first_poll()) / 1000
@@ -298,7 +391,8 @@ class Client(object):
         r = resp.query_execute()
         progress = r.get_progress()
         callback(r, progress)
-        while r.get_status() == Client.STATUS_RUNNING:
+        while r.get_status() == Client.STATUS_RUNNING \
+                and r.get_http_status() == 200:
             sleep(np)
             # Running -> Running | Finished | Failed
             r = resp.query_execute()
@@ -311,7 +405,7 @@ class Client(object):
            The service must run in 'async' mode"""
 
         if suid is not None:
-            # skip ckecking the async mode
+            # skip checking the async mode
             if not self.is_async():
                 raise Exception("not async.")
         service = self.get_service_url()
@@ -319,15 +413,15 @@ class Client(object):
             raise Exception("no service_url.")
         a = re.search("/[md]/", service)
         if a is None:
-            raise Exception("not a model/data service_url.")
+            raise Exception("not a model/data service_url" + service)
 
         suid = suid or self.get_suid()
         if suid is None:
             raise Exception("no suid.")
         query_url = service[:a.start()] + "/q/" + suid
         # print(query_url)
-        c = Client.__get(query_url, section=self.RESULT)
-        c.parent = self
+        c = Client.__get(query_url, section=self.RESULT, conf=self.conf)
+        c.parent = self.parent
         return c
 
     #### files handling
@@ -367,8 +461,8 @@ class Client(object):
                 raise Exception("Not in data: " + name)
             val = self.get_data_value(name)
             print("  --> " + name)
-            r = requests.get(val, allow_redirects=True)
-            open(os.path.join(dir, name), "wb").write(r.content)
+            response = Client.__get0(val, self.conf)
+            open(os.path.join(dir, name), "wb").write(response.content)
 
     #### pubsub related properties
 
@@ -419,6 +513,16 @@ class Client(object):
         """Set the delay between p/s submissions in ms"""
         self._delay = d
 
+    @property
+    def config(self) -> Dict:
+        """ ??? """
+        return self.conf
+
+    @delay.setter
+    def config(self, c: Dict) -> None:
+        """ ??? """
+        self.conf = c
+
     #### metainfo
 
     def get_metainfo(self, name: str) -> str:
@@ -476,7 +580,8 @@ class Client(object):
             The entry must exist"""
         self.data[name][self.KEY_VALUE] = value
 
-    def add_data(self, name, value=None, descr: str = None, unit: str = None, geometry_type: str = None, coord: [] = None) -> None:
+    def add_data(self, name, value=None, descr: str = None, unit: str = None, geometry_type: str = None,
+                 coord: [] = None) -> None:
         """Add a new entry to the data"""
         if isinstance(name, Dict):
             # name is the the json data object 
@@ -571,12 +676,21 @@ class Client(object):
         """Check if execution mode is async """
         return self.metainfo.get(self.META_KEY_MODE, self.META_VAL_SYNC) == self.META_VAL_ASYNC
 
+    def is_finished(self) -> bool:
+        """Check if execution mode is async """
+        return self.http_status == 200 and self.metainfo.get(
+            self.META_KEY_STATUS, self.STATUS_UNKNOWN) == self.STATUS_FINISHED
+
     def get_first_poll(self) -> int:
         return self.metainfo.get(self.META_KEY_FIRST_POLL, 2000)
 
     def get_next_poll(self) -> int:
         return self.metainfo.get(self.META_KEY_NEXT_POLL, 2000)
 
+    #### HTTP
+    def get_http_status(self) -> int:
+        return self.http_status
+
     #### printing support
 
     def get_entry_asstr(self, name) -> str:
@@ -605,8 +719,9 @@ class Client(object):
     def __str__(self):
         """String representation."""
         return (
-            "SERVICE: '{name}' \n url: {url}\n description: {description}\n parent: {parentsuid} \n metainfo: {metainfo}\n data:\n{data}".format(
+            "SERVICE: '{name}' \n url: {url}\n description: {description}\n parent: {parentsuid}\n http_status: {http} \n metainfo: {metainfo}\n data:\n{data}".format(
                 name=self.get_name(), suid=self.get_suid(), url=self.get_url(),
                 description=self.get_description(),
                 parentsuid=(self.parent.get_suid() if self.parent is not None else 'None'),
+                http=self.get_http_status(),
                 metainfo=self.metainfo_tostr(indent=2).replace("\n", "\n           "), data=self.get_entries_asstr()))
